from math import log
import torch
import torch.nn as nn
from torchvision.ops.boxes import box_iou
from tasks.config import cfg
import torchvision.ops as ops
import torch.nn.functional as F

DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


def mil_loss(cls_score, labels):
    cls_score = cls_score.clamp(1e-6, 1 - 1e-6)
    labels = labels.clamp(0, 1)
    loss = -labels * torch.log(cls_score) - (1 - labels) * torch.log(1 - cls_score)

    return loss.mean()


def L2_loss(cls_score, labels):
    cls_score = cls_score.clamp(1e-6, 1 - 1e-6)
    labels = labels.clamp(0, 1)
    loss = (cls_score - labels) ** 2
    return loss.mean()


def focal_mil_loss(cls_score, labels, alpha=0.25, gamma=2):
    cls_score = cls_score.clamp(1e-6, 1 - 1e-6)
    labels = labels.clamp(0, 1)
    raw_loss = F.binary_cross_entropy(cls_score, labels, reduction='none')
    pt = torch.exp(- raw_loss)  # prevents nans when probability 0
    focal_loss = alpha * (1 - pt) ** gamma * raw_loss
    return focal_loss.mean()


class OICRLosses(nn.Module):
    def forward(self, pcl_probs, labels, cls_loss_weights, gt_assignment, im_labels):
        eps = 1e-6
        pcl_probs = pcl_probs.clamp(eps, 1 - eps).log()
        cls_loss_weights = cls_loss_weights.repeat(pcl_probs.shape[1], 1).permute(1, 0).cuda()

        labels = labels.repeat(pcl_probs.shape[1], 1).permute(1, 0).long()
        reap = torch.arange(pcl_probs.shape[1])[None, :].repeat(pcl_probs.shape[0], 1).long()
        labels = (reap - labels == 0).float().to(DEVICE)

        loss = labels * cls_loss_weights * pcl_probs
        loss = - loss.sum(dim=0).sum() / pcl_probs.size(0)
        return loss


def Refinement(boxes, cls_prob, im_labels, lambda_gt=0.5, lambda_ign=-1.0):
    im_labels = im_labels.long()

    cls_prob = cls_prob.clone().detach()

    # if cls_prob have the background dimension, we cut it out
    if cls_prob.shape[1] != im_labels.shape[1]:
        cls_prob = cls_prob[:, 1:]

    #  avoiding NaNs.
    eps = 1e-9
    cls_prob = cls_prob.clamp(eps, 1 - eps)

    num_images, num_classes = im_labels.shape
    max_values, max_indexes = cls_prob.max(dim=0)  # select the proposal with highest score from (k-1)th OICR

    gt_boxes = boxes[max_indexes, :][im_labels[0] == 1, :]  # treat the proposal with highest score as ground truth
    gt_classes = torch.arange(num_classes)[im_labels[0] == 1].view(-1, 1) + 1
    gt_scores = max_values[im_labels[0] == 1].view(-1, 1)

    overlaps = box_iou(boxes, gt_boxes)  # calculate IOU
    max_overlaps, gt_assignment = overlaps.max(dim=1)  # max IOU and Index
    ori_gt_assignment = gt_assignment.clone()
    labels = gt_classes[gt_assignment, 0]
    cls_loss_weights = gt_scores[gt_assignment, 0]

    bg_inds = torch.where(max_overlaps < lambda_gt)[0]
    # ig_inds = torch.where(max_overlaps < lambda_ign)[0]
    # cls_loss_weights[ig_inds] = 0.0
    labels[bg_inds] = 0
    gt_assignment[bg_inds] = -1

    return {'labels': labels.reshape(1, -1),
            'cls_loss_weights': cls_loss_weights.reshape(1, -1),
            'gt_assignment': gt_assignment.reshape(1, -1),
            'im_labels_real': torch.cat((torch.tensor([[1]]).cuda(), im_labels), dim=1),
            'gt_boxes': gt_boxes,
            'ori_gt_assignment': ori_gt_assignment.reshape(1, -1),
            }


class RoiPoolLayer(nn.Module):
    def __init__(self, dim_in, spatial_scale):
        super().__init__()
        self.dim_in = dim_in

        res = cfg.FAST_RCNN.ROI_XFORM_RESOLUTION
        self.roi_pool = ops.RoIPool(output_size=(res, res), spatial_scale=spatial_scale)

        self.spatial_scale = spatial_scale
        self.dim_out = hidden_dim = 4096

        roi_size = cfg.FAST_RCNN.ROI_XFORM_RESOLUTION
        self.fc1 = nn.Linear(dim_in * roi_size ** 2, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)

    def detectron_weight_mapping(self):
        detectron_weight_mapping = {
            'fc1.weight': 'fc6_w',
            'fc1.bias': 'fc6_b',
            'fc2.weight': 'fc7_w',
            'fc2.bias': 'fc7_b'
        }
        return detectron_weight_mapping, []

    def forward(self, x, rois):
        x = self.roi_pool(x, rois)  # [bs,512,7,7]
        batch_size = x.size(0)
        x = F.relu(self.fc1(x.view(batch_size, -1)), inplace=True)  # [bs,4096]
        x = F.relu(self.fc2(x), inplace=True)  # [bs,4096]

        return x


class BoxRegressor(nn.Module):
    def __init__(self, dim_in):
        super().__init__()
        self.regression_fc = nn.Linear(dim_in, 4)
        self._init_weights()

    def _init_weights(self):
        nn.init.normal_(self.regression_fc.weight, std=0.01)
        nn.init.constant_(self.regression_fc.bias, 0.01)

    def detectron_weight_mapping(self):
        return dict(), []

    def forward(self, x):
        return torch.sigmoid(self.regression_fc(x))


def bbox_regressor_transformation(original_boxes, offsets, max_width=None, max_height=None):
    """
    refer to "Fast-RCNN"
    :param original_boxes:      torch[n x 4] (x_min, y_min, x_max, y_max)
    :param offsets:             torch[n x 4] (dx, dy, dw, dh)
    :param max_height:
    :param max_width:
    :return: transformed_boxes: torch[n x 4] (x_min, y_min, x_max, y_max)
    """

    mx = (original_boxes[:, 0] + original_boxes[:, 2]) / 2.0
    my = (original_boxes[:, 1] + original_boxes[:, 3]) / 2.0
    pw = original_boxes[:, 2] - original_boxes[:, 0]
    ph = original_boxes[:, 3] - original_boxes[:, 1]
    scaled_w = pw * torch.exp(offsets[:, 2])
    scaled_h = ph * torch.exp(offsets[:, 3])

    translated_mx = mx + offsets[:, 0] * pw
    translated_my = my + offsets[:, 1] * ph

    transformed_xmin = translated_mx - scaled_w / 2.0
    transformed_ymin = translated_my - scaled_h / 2.0
    transformed_xmax = translated_mx + scaled_w / 2.0
    transformed_ymax = translated_my + scaled_h / 2.0

    if max_width is None:
        max_width = torch.max(original_boxes[:, 2])

    if max_height is None:
        max_height = torch.max(original_boxes[:, 3])

    transformed_xmin[transformed_xmin < 0] = 0.0
    transformed_xmin[transformed_xmin > max_width] = float(max_width)

    transformed_ymin[transformed_ymin < 0] = 0.0
    transformed_ymin[transformed_ymin > max_height] = float(max_height)

    transformed_xmax[transformed_xmax < 0] = 0.0
    transformed_xmax[transformed_xmax > max_width] = float(max_width)

    transformed_ymax[transformed_ymax < 0] = 0.0
    transformed_ymax[transformed_ymax > max_height] = float(max_height)

    return torch.stack((transformed_xmin, transformed_ymin, transformed_xmax, transformed_ymax)).transpose(0, 1)


def get_gt_offsets(gt_boxes, proposals, ori_gt_assignment):
    """
    refer to "Rich feature hierarchies for accurate object detection and semantic segmentation"
    :param gt_boxes:  torch[n x 4] (x_min, y_min, x_max, y_max)
    :param proposals: torch[m x 4] (x_min, y_min, x_max, y_max)
    :return:
    """
    # assert (len(gt_boxes.shape) == 2 and gt_boxes.shape[1] == 4)
    # assert (len(proposals.shape) == 2 and proposals.shape[1] == 4)

    gt_boxes = gt_boxes[ori_gt_assignment.squeeze(), :]
    # assert (gt_boxes.shape[0] == proposals.shape[0])

    gt_mx = (gt_boxes[:, 0] + gt_boxes[:, 2]) / 2.0
    gt_my = (gt_boxes[:, 1] + gt_boxes[:, 3]) / 2.0
    gt_w, gt_h = gt_boxes[:, 2] - gt_boxes[:, 0], gt_boxes[:, 3] - gt_boxes[:, 1]

    p_mx = (proposals[:, 0] + proposals[:, 2]) / 2.0
    p_my = (proposals[:, 1] + proposals[:, 3]) / 2.0
    pw, ph = proposals[:, 2] - proposals[:, 0], proposals[:, 3] - proposals[:, 1]

    tx = (gt_mx - p_mx) / pw
    ty = (gt_my - p_my) / ph
    tw = torch.log(gt_w / pw)
    th = torch.log(gt_h / ph)

    return torch.stack((tx, ty, tw, th)).transpose(0, 1)


def box_regression_loss(gt_offsets, pred_offsets, assignment, scores):
    # print('gt_boxes', gt_boxes.shape)
    # print('preds', preds.shape)
    # print('assignment', assignment.shape)
    # print('scores', scores.shape)
    _assignment_mask = (assignment >= 0).squeeze()
    # positive_index = torch.where(assignment >= 0)[1]
    # print('positive_index', positive_index.shape)
    _gt_offsets = gt_offsets[_assignment_mask, :]
    _pred_offsets = pred_offsets[_assignment_mask, :]
    # print('gt_boxes', gt_boxes.shape)
    # assignment = assignment[positive_index, :]
    # print('assignment', assignment.shape)
    # print('SHAPE VIEW:', gt_boxes.shape, preds.shape)
    smooth_l1_values = F.smooth_l1_loss(_gt_offsets, _pred_offsets, reduction='none')
    # print('smooth_l1_values', smooth_l1_values.shape)
    return torch.mean(torch.sum(smooth_l1_values * scores[_assignment_mask, 1:], dim=1))


class WrapperBatchNorm(nn.Module):
    def __init__(self, dim_in):
        super().__init__()
        self.norm_wrapper = nn.BatchNorm2d(dim_in)

    def detectron_weight_mapping(self):
        return dict(), []

    def forward(self, x):
        return self.norm_wrapper(x)


def get_adaptative_lambda(inner_iter, max_iter):
    lb = 100
    lambda_gt = (log(inner_iter + lb) - log(lb)) / (log(max_iter + lb) - log(lb)) / 2
    lambda_ign = 0.51 - lambda_gt

    return lambda_gt, lambda_ign


class Distillation(nn.Module):

    def __init__(self, dim_in, dim_out):
        super().__init__()

        self.distillation = nn.Linear(dim_in, dim_out)
        self._init_weights()

    def _init_weights(self):
        nn.init.normal_(self.distillation.weight, std=0.01)
        nn.init.constant_(self.distillation.bias, 0)

    def detectron_weight_mapping(self):
        return dict(), []

    def forward(self, x):
        if x.dim() == 4:
            x = x.squeeze(3).squeeze(2)

        distillation = F.softmax(self.distillation(x), dim=1)

        return distillation
